#include "main.h"
#include "font.h"

/* Private defines */
#define FRAMEBUFFER_ROTATE

/* Private constants */
#define MAX_COLUMN    (DISPLAY_COLUMNS - 1)
#define FONT_WIDTH    8

/* private macros */
/* Private variables */
uint8_t display_Buffer[DISPLAY_COLUMNS] = {0};

/* Privae fuctions */
static void _display_WriteBits(uint16_t data, uint16_t nbits);
static void _delay_c(uint32_t cycle);

/* Exported fuctions */

/**
 * @brief Initialization HT1632C
 */
void display_Init(void) {
  /* Wait for SPI */
  while ((SPI1->SR & SPI_SR_BSY) != 0) { __NOP(); }

  GPIO_SPI_SW();
  HT1632C_CS_ON;

  // Turn on system oscillator
  _display_WriteBits(0x802, 0x800);
  // Turn on LED duty cycle generator
  _display_WriteBits(0x6, 0x100);

  HT1632C_CS_OFF;
  GPIO_SPI_HW();

  display_Fill(0x0);
}

/**
 * @brief Out ot display one symbol in specified column
 * 
 * @param symb Symbol to display
 * @param column column from 0 to DISPLAY_COLUMNS-1
 */
void display_Char(const uint8_t symb, const uint8_t column) {
  int i=0, c = column;

  if (c > MAX_COLUMN) {
    c = MAX_COLUMN;
  }

  while ((i < FONT_WIDTH) && (c < DISPLAY_COLUMNS)) {
    display_Buffer[c] = Font_6x8[symb][i];
    i ++;
    c ++;
  }
}

/**
 * @brief Set HT1632C PWM Value
 * @param pwm value in 0-15
 */
void display_PWM(uint8_t pwm) {
  // check value
  if (pwm > 15) {
    pwm = 15;
  }
  // align value
  pwm <<= 1;

  /* Wait for SPI */
  while ((SPI1->SR & SPI_SR_BSY) != 0) { __NOP(); }

  GPIO_SPI_SW();

  HT1632C_CS_ON;
  _display_WriteBits((0x940|pwm), 0x800);
  HT1632C_CS_OFF;

  GPIO_SPI_HW();
}

/**
 * @brief Set HT1632C Blink state
 * @param state 0 for off, any other value for on
 */
void display_BlinkState(dis_en_t state) {
  /* Wait for SPI */
  while ((SPI1->SR & SPI_SR_BSY) != 0) { __NOP(); }

  GPIO_SPI_SW();

  HT1632C_CS_ON;
  if (state == Disable) {
    _display_WriteBits(0x810, 0x800); // Blink off
  } else {
    _display_WriteBits(0x812, 0x800); // Blink on
  }
  HT1632C_CS_OFF;

  GPIO_SPI_HW();
}

/**
 * @brief Set HT1632C Led state
 * @param state 0 for off, any other value for on
 */
void display_LedState(dis_en_t state) {
  /* Wait for SPI */
  while ((SPI1->SR & SPI_SR_BSY) != 0) { __NOP(); }

  GPIO_SPI_SW();

  HT1632C_CS_ON;
  if (state == Disable) {
    _display_WriteBits(0x804, 0x800); // Led off
  } else {
    _display_WriteBits(0x806, 0x800); // Led on
  }
  HT1632C_CS_OFF;

  GPIO_SPI_HW();
}

/**
 * @brief Primitive delay
 */
static void _delay_c(uint32_t cycle) {
  while (cycle) {
    __NOP();
    cycle --;
  }
}

/**
 * @brief Write <nbits> bit of data to selected HT1632Cs
 * @param data words to write
 * @param nbits num of bits (1<<(num-1))
 */
void _display_WriteBits(const uint16_t data, uint16_t nbits) {
  _delay_c(2);
  do {
    HT1632C_WR_LOW;

    if (data & nbits) {
      HT1632C_DATA_1;
    } else {
      HT1632C_DATA_0;
    }
    _delay_c(5);
    HT1632C_WR_HGH;
    _delay_c(5);
  } while (nbits >>= 1);
}

/**
 * @brief Write Data to display buffer, no transfer to HT1632.
 * @param data pointer to data array
 * @param addr begin address
 * @param len bytes to write
 */
void display_WriteData(const uint8_t * data, uint8_t addr, uint8_t len) {
  /* check given values */
  if (len == 0) {
    return;
  }
  if (addr >= DISPLAY_COLUMNS) {
    addr = DISPLAY_COLUMNS - 1;
  }
  if ((addr + len) > DISPLAY_COLUMNS) {
    len = DISPLAY_COLUMNS - addr;
  }

#ifdef FRAMEBUFFER_ROTATE
  /* Copy given data to framebuffer */
  int i, a;
  for (i=0; i<len; i++) {
    a = addr + i;
    display_Buffer[a] = data[i];
  }
  //display_WriteBuffer();

#else
  // without dma - pool mode
  int i;

  GPIO_SPI_SW();
  HT1632C_CS_ON;
  i = (0x280 | addr);
  // write addr
  _display_WriteBits(i, 0x200);

  GPIO_SPI_HW();
  while ((SPI1->SR & SPI_SR_BSY) != 0) { __NOP(); }
  for (i=0; i<len; i++) {
    while ((SPI1->SR & SPI_SR_TXE) == 0) { __NOP(); }
    SPI1->DR = data[i];
  }
  // wait for end of transmission
  while ((SPI1->SR & SPI_SR_BSY) != 0) { __NOP(); }
  HT1632C_CS_OFF;

  // with dam
  uint16_t a = 0x280 | addr;

  /* Wait for SPI */
  while ((SPI1->SR & SPI_SR_BSY) != 0) { __NOP(); }

  /* DMA Source addr: Address of the SPI buffer. */
  DMA1_Channel3->CMAR = data;
  /* Set DMA data transfer length (SPI buffer length). */
  DMA1_Channel3->CNDTR = len;

  GPIO_SPI_SW();

  HT1632C_CS_ON;
  _display_WriteBits(a, 0x200);

  GPIO_SPI_HW();

  // start transfer
  SPI1->CR2 |= (SPI_CR2_TXDMAEN);
  DMA1_Channel3->CCR |= DMA_CCR_EN;
  // End of transaction in DMA1_Channel2_3_IRQHandler()
#endif /* FRAMEBUFFER_ROTATE */
}

/**
 * @brief Write all buffer to HT1632C
 */
void display_WriteBuffer(void) {
  /* Wait for SPI */
  while ((SPI1->SR & SPI_SR_BSY) != 0) { __NOP(); }

#ifdef FRAMEBUFFER_ROTATE
  uint8_t spi_buf[DISPLAY_COLUMNS] = {0};

  /* Rotate display buffer */
  uint8_t i, j, p = 0x01;
  for (i=0; i<8; i++) {
    if ((display_Buffer[0] & p) != 0) { spi_buf[i] |= 0x01; }
    if ((display_Buffer[1] & p) != 0) { spi_buf[i] |= 0x02; }
    if ((display_Buffer[2] & p) != 0) { spi_buf[i] |= 0x04; }
    if ((display_Buffer[3] & p) != 0) { spi_buf[i] |= 0x08; }
    if ((display_Buffer[4] & p) != 0) { spi_buf[i] |= 0x10; }
    if ((display_Buffer[5] & p) != 0) { spi_buf[i] |= 0x20; }
    if ((display_Buffer[6] & p) != 0) { spi_buf[i] |= 0x40; }
    if ((display_Buffer[7] & p) != 0) { spi_buf[i] |= 0x80; }

    j = i + 8;
    if ((display_Buffer[8] & p) != 0)  { spi_buf[j] |= 0x01; }
    if ((display_Buffer[9] & p) != 0)  { spi_buf[j] |= 0x02; }
    if ((display_Buffer[10] & p) != 0) { spi_buf[j] |= 0x04; }
    if ((display_Buffer[11] & p) != 0) { spi_buf[j] |= 0x08; }
    if ((display_Buffer[12] & p) != 0) { spi_buf[j] |= 0x10; }
    if ((display_Buffer[13] & p) != 0) { spi_buf[j] |= 0x20; }
    if ((display_Buffer[14] & p) != 0) { spi_buf[j] |= 0x40; }
    if ((display_Buffer[15] & p) != 0) { spi_buf[j] |= 0x80; }

    j = i + 16;
    if ((display_Buffer[16] & p) != 0) { spi_buf[j] |= 0x01; }
    if ((display_Buffer[17] & p) != 0) { spi_buf[j] |= 0x02; }
    if ((display_Buffer[18] & p) != 0) { spi_buf[j] |= 0x04; }
    if ((display_Buffer[19] & p) != 0) { spi_buf[j] |= 0x08; }
    if ((display_Buffer[20] & p) != 0) { spi_buf[j] |= 0x10; }
    if ((display_Buffer[21] & p) != 0) { spi_buf[j] |= 0x20; }
    if ((display_Buffer[22] & p) != 0) { spi_buf[j] |= 0x40; }
    if ((display_Buffer[23] & p) != 0) { spi_buf[j] |= 0x80; }

    p <<= 1;
  }

  /* DMA Source addr: Address of the SPI buffer. */
  DMA1_Channel3->CMAR = (uint32_t)&spi_buf[0];
#else
  /* DMA Source addr: Address of the SPI buffer. */
  DMA1_Channel3->CMAR = (uint32_t)&display_Buffer[0];
#endif /* FRAMEBUFFER_ROTATE */

  /* Set DMA data transfer length (SPI buffer length). */
  DMA1_Channel3->CNDTR = DISPLAY_COLUMNS;

  GPIO_SPI_SW();

  HT1632C_CS_ON;
  _display_WriteBits(0x280, 0x200);

  GPIO_SPI_HW();

  /* start transfer */
  SPI1->CR2 |= SPI_CR2_TXDMAEN;
  DMA1_Channel3->CCR |= DMA_CCR_EN;
  /* End of transaction in DMA1_Channel2_3_IRQHandler() */
}

/**
 * @brief Fill display.
 * @param pattern for filling
 */
void display_Fill(uint8_t pattern) {
  int i;
  for (i=0; i<DISPLAY_COLUMNS; i++) {
    display_Buffer[i] = 0x0;
  }
  display_WriteBuffer();
}
